from __future__ import annotations

from typing import TYPE_CHECKING

from openai import APIError, AsyncOpenAI

from shelloracle.providers import Provider, ProviderError, Setting, system_prompt

if TYPE_CHECKING:
    from collections.abc import AsyncIterator


class XAI(Provider):
    name = "XAI"

    api_key = Setting(default="")
    model = Setting(default="grok-beta")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if not self.api_key:
            msg = "No API key provided"
            raise ProviderError(msg)
        self.client = AsyncOpenAI(
            api_key=self.api_key,
            base_url="https://api.x.ai/v1",
        )

    async def generate(self, prompt: str) -> AsyncIterator[str]:
        try:
            stream = await self.client.chat.completions.create(
                model=self.model,
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": prompt},
                ],
                stream=True,
            )
            async for chunk in stream:
                if chunk.choices[0].delta.content is not None:
                    yield chunk.choices[0].delta.content
        except APIError as e:
            msg = f"Something went wrong while querying XAI: {e}"
            raise ProviderError(msg) from e



from __future__ import annotations

import asyncio
import logging
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from prompt_toolkit import PromptSession
from prompt_toolkit.application import create_app_session_from_tty
from prompt_toolkit.history import FileHistory
from prompt_toolkit.patch_stdout import patch_stdout
from yaspin import yaspin
from yaspin.spinners import Spinners

from shelloracle.providers import get_provider

if TYPE_CHECKING:
    from yaspin.core import Yaspin

    from shelloracle.cli import Application

logger = logging.getLogger(__name__)


async def prompt_user(default_prompt: str | None = None) -> str:
    # stdin doesn't exist when running as a zle widget
    with create_app_session_from_tty(), patch_stdout():
        history_file = Path.home() / ".shelloracle_history"
        prompt_session: PromptSession = PromptSession(history=FileHistory(str(history_file)))
        prompt_session.output.write_raw("\033[E")
        return await prompt_session.prompt_async("> ", default=default_prompt or "")


def get_query_from_pipe() -> str | None:
    """Get a query from stdin pipe.

    :raises ValueError: If the input is more than one line
    :return: The query from the stdin pipe
    """
    if os.isatty(0) or not (lines := sys.stdin.readlines()):  # Return 'None' if fd 0 is a tty (no pipe)
        return None
    if len(lines) > 1:
        msg = "Multi-line input is not supported"
        raise ValueError(msg)
    logger.debug("using query from stdin: %s", lines)
    return lines[0].rstrip()


def spinner(style: str | None) -> Yaspin:
    """Get the correct spinner based on the user's configuration

    :param style: The spinner style
    :returns: yaspin object
    """
    if style:
        style = getattr(Spinners, style)
        return yaspin(style)
    return yaspin()


async def shelloracle(app: Application) -> None:
    """ShellOracle program entrypoint

    If there is a query from the input pipe, it processes the query to generate a response.
    If there isn't a query from the input pipe, it prompts the user for input.

    Environment variables:
        - SHOR_DEFAULT_PROMPT: This is the initial user prompt that can be configured via this environment variable.

    :returns: None
    :raises KeyboardInterrupt: if the user presses CTRL+C
    """
    if not (prompt := get_query_from_pipe()):
        default_prompt = os.environ.get("SHOR_DEFAULT_PROMPT")
        prompt = await prompt_user(default_prompt)
    logger.info("user prompt: %s", prompt)

    provider = get_provider(app.configuration.provider)(app.configuration)

    shell_command = ""
    with create_app_session_from_tty(), patch_stdout(raw=True), spinner(app.configuration.spinner_style) as sp:
        async for token in provider.generate(prompt):
            # some models may erroneously return a newline, which causes issues with the status spinner
            shell_command += token.replace("\n", "")
            sp.text = shell_command
    logger.info("generated shell command: %s", shell_command)
    sys.stdout.write(shell_command)


def cli(app: Application) -> None:
    """Run the ShellOracle command line interface

    :returns: None
    """
    try:
        asyncio.run(shelloracle(app))
    except (KeyboardInterrupt, asyncio.exceptions.CancelledError):
        return
    except Exception:
        logger.exception("An error occurred")



from __future__ import annotations

import logging
import sys
from collections.abc import Iterator, Mapping
from typing import TYPE_CHECKING, Any

from yaspin.spinners import SPINNERS_DATA

if TYPE_CHECKING:
    from pathlib import Path


if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

logger = logging.getLogger(__name__)


class Configuration(Mapping):
    def __init__(self, config: dict[str, Any]) -> None:
        """ShellOracle application configuration

        :param config: configuration dict
        :raises FileNotFoundError: if the configuration file does not exist
        """
        self._config = config

    def __getitem__(self, key: str) -> Any:
        return self._config[key]

    def __len__(self) -> int:
        return len(self._config)

    def __iter__(self) -> Iterator[Any]:
        return iter(self._config)

    def __str__(self):
        return f"Configuration({self._config})"

    def __repr__(self) -> str:
        return str(self)

    @property
    def raw_config(self) -> dict[str, Any]:
        return self._config

    @property
    def provider(self) -> str:
        return self["shelloracle"]["provider"]

    @property
    def spinner_style(self) -> str | None:
        style = self["shelloracle"].get("spinner_style", None)
        if not style:
            return None
        if style not in SPINNERS_DATA:
            logger.warning("invalid spinner style: %s", style)
            return None
        return style

    @classmethod
    def from_file(cls, filepath: Path):
        with filepath.open("rb") as config_file:
            config = tomllib.load(config_file)
        return cls(config)



from __future__ import annotations

import pytest
import tomli_w

from shelloracle.config import Configuration


class TestConfiguration:
    @pytest.fixture
    def default_config(self):
        return Configuration(
            {
                "shelloracle": {"provider": "Ollama", "spinner_style": "earth"},
                "provider": {"Ollama": {"host": "localhost", "port": 11434, "model": "dolphin-mistral"}},
            }
        )

    def test_from_file(self, default_config, tmp_path):
        config_path = tmp_path / "config.toml"
        with config_path.open("wb") as f:
            tomli_w.dump(default_config.raw_config, f)
        assert Configuration.from_file(config_path) == default_config

    def test_getitem(self, default_config):
        for key in default_config:
            assert default_config[key] == default_config.raw_config[key]

    def test_len(self, default_config):
        assert len(default_config) == len(default_config.raw_config)

    def test_iter(self, default_config):
        assert list(iter(default_config)) == list(iter(default_config.raw_config))

    def test_provider(self, default_config):
        assert default_config.provider == "Ollama"

    def test_spinner_style(self, default_config):
        assert default_config.spinner_style == "earth"

    def test_no_spinner_style(self, caplog):
        config = Configuration(
            {
                "shelloracle": {"provider": "Ollama"},
                "provider": {"Ollama": {"host": "localhost", "port": 11434, "model": "dolphin-mistral"}},
            }
        )
        assert config.spinner_style is None
        assert "invalid spinner style" not in caplog.text

    def test_invalid_spinner_style(self, caplog):
        config = Configuration(
            {
                "shelloracle": {"provider": "Ollama", "spinner_style": "invalid"},
                "provider": {"Ollama": {"host": "localhost", "port": 11434, "model": "dolphin-mistral"}},
            }
        )
        assert config.spinner_style is None
        assert "invalid spinner style" in caplog.text



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from shelloracle.cli import Application


@click.command()
@click.pass_obj
def init(app: Application):
    """Install shelloracle keybindings."""
    # nest this import in a function to avoid expensive module loads
    from shelloracle.bootstrap import bootstrap_shelloracle

    bootstrap_shelloracle(app.config_path)



from __future__ import annotations

import logging
import sys
from typing import TYPE_CHECKING

import click

from shelloracle import shelloracle
from shelloracle.cli.application import Application
from shelloracle.cli.config import config
from shelloracle.config import Configuration
from shelloracle.tty_log_handler import TtyLogHandler

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


def configure_logging(log_path: Path):
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    file_formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(message)s")
    file_handler = logging.FileHandler(log_path)
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(file_formatter)

    tty_formatter = logging.Formatter("%(message)s")
    tty_handler = TtyLogHandler()
    tty_handler.setLevel(logging.WARNING)
    tty_handler.setFormatter(tty_formatter)

    root_logger.addHandler(file_handler)
    root_logger.addHandler(tty_handler)


@click.group(invoke_without_command=True)
@click.version_option()
@click.pass_context
def cli(ctx: click.Context):
    """ShellOracle command line interface."""
    app = Application()
    configure_logging(app.log_path)
    ctx.obj = app

    if ctx.invoked_subcommand is not None:
        # If no subcommand is invoked, run the main CLI
        return

    try:
        app.configuration = Configuration.from_file(app.config_path)
    except FileNotFoundError:
        logger.warning("Configuration not found. Run `shor config init` to initialize.")
        sys.exit(1)

    shelloracle.cli(app)


cli.add_command(config)


def main():
    cli()



from __future__ import annotations

import logging

from prompt_toolkit import print_formatted_text
from prompt_toolkit.application import create_app_session_from_tty
from prompt_toolkit.formatted_text import FormattedText


class TtyLogHandler(logging.Handler):
    def emit(self, record: logging.LogRecord):
        if record.levelno >= logging.ERROR:
            color = "ansired"
        elif record.levelno == logging.WARNING:
            color = "ansiyellow"
        else:
            color = "ansywhite"
        log_entry = self.format(record)
        formatted_log_entry = FormattedText([(color, f"\n{log_entry}")])
        with create_app_session_from_tty():
            print_formatted_text(formatted_log_entry)



from __future__ import annotations

import typing
from pathlib import Path

if typing.TYPE_CHECKING:
    from shelloracle.config import Configuration

shelloracle_home = Path.home() / ".shelloracle"
shelloracle_home.mkdir(exist_ok=True)


class Application:
    configuration: Configuration

    def __init__(self):
        self.config_path = shelloracle_home / "config.toml"
        self.log_path = shelloracle_home / "shelloracle.log"



import pytest

from shelloracle.config import Configuration
from shelloracle.providers.deepseek import Deepseek


class TestOpenAI:
    @pytest.fixture
    def deepseek_config(self):
        config = {
            "shelloracle": {"provider": "Deepseek"},
            "provider": {
                "Deepseek": {
                    "api_key": "sk-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
                    "model": "grok-beta",
                }
            },
        }
        return Configuration(config)

    @pytest.fixture
    def deepseek_instance(self, deepseek_config):
        return Deepseek(deepseek_config)

    def test_name(self):
        assert Deepseek.name == "Deepseek"

    def test_api_key(self, deepseek_instance):
        assert (
            deepseek_instance.api_key
            == "sk-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
        )

    def test_model(self, deepseek_instance):
        assert deepseek_instance.model == "grok-beta"

    @pytest.mark.asyncio
    async def test_generate(self, mock_asyncopenai, deepseek_instance):
        result = ""
        async for response in deepseek_instance.generate(""):
            result += response
        assert result == "head -c 100 /dev/urandom | hexdump -C"



from __future__ import annotations

from typing import TYPE_CHECKING

import google.generativeai as genai

from shelloracle.providers import Provider, ProviderError, Setting, system_prompt

if TYPE_CHECKING:
    from collections.abc import AsyncIterator


class Google(Provider):
    name = "Google"

    api_key = Setting(default="")
    model = Setting(default="gemini-2.0-flash")  # Assuming a default model name

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if not self.api_key:
            msg = "No API key provided"
            raise ProviderError(msg)
        genai.configure(api_key=self.api_key)
        self.model_instance = genai.GenerativeModel(self.model, system_instruction=system_prompt)

    async def generate(self, prompt: str) -> AsyncIterator[str]:
        try:
            response = await self.model_instance.generate_content_async(
                [prompt],
                stream=True,
            )

            async for chunk in response:
                yield chunk.text
        except Exception as e:
            msg = f"Something went wrong while querying Google Gemini: {e}"
            raise ProviderError(msg) from e



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from shelloracle.cli.application import Application


@click.command()
@click.pass_obj
def edit(app: Application):
    """Edit shelloracle configuration."""
    click.edit(filename=app.config_path)



import pytest

from shelloracle.config import Configuration
from shelloracle.providers.xai import XAI


class TestOpenAI:
    @pytest.fixture
    def xai_config(self):
        config = {
            "shelloracle": {"provider": "XAI"},
            "provider": {
                "XAI": {
                    "api_key": "xai-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
                    "model": "grok-beta",
                }
            },
        }
        return Configuration(config)

    @pytest.fixture
    def xai_instance(self, xai_config):
        return XAI(xai_config)

    def test_name(self):
        assert XAI.name == "XAI"

    def test_api_key(self, xai_instance):
        assert (
            xai_instance.api_key
            == "xai-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
        )

    def test_model(self, xai_instance):
        assert xai_instance.model == "grok-beta"

    @pytest.mark.asyncio
    async def test_generate(self, mock_asyncopenai, xai_instance):
        result = ""
        async for response in xai_instance.generate(""):
            result += response
        assert result == "head -c 100 /dev/urandom | hexdump -C"
